/*------------------------------------------------------------------------------------------*\
 *   Copyright (C) 2012 AVM GmbH <fritzbox_info@avm.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
\*------------------------------------------------------------------------------------------*/


#include <stdarg.h>
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/skbuff.h>
#include <linux/err.h>
#include <linux/delay.h>
#include <linux/netdevice.h>
#include <net/pkt_sched.h>
#include <avmnet_debug.h>
#include <avmnet_module.h>

#if defined(CONFIG_IP_MULTICAST_FASTFORWARD)
/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static void avmnet_mcfw_handle_tasklet(unsigned long data) {
    avmnet_device_t *avm_dev = (avmnet_device_t *) data;
    avmnet_netdev_priv_t *priv = netdev_priv(avm_dev->device);
    struct sk_buff *skb;
    int i = 32;

    while(i-- > 0 && (skb = skb_dequeue(&priv->mcfw_queue)) != 0) {
        dev_queue_xmit(skb);
    }
    if(skb_queue_len(&priv->mcfw_queue)) {
        tasklet_schedule(&priv->mcfw_tasklet);
    }
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static void avmnet_dev_queue_xmit(avmnet_device_t *avm_dev, struct sk_buff *skb) {
    struct ethhdr *eth = (struct ethhdr *) skb->data;
    struct net_device *dev = avm_dev->device;
    avmnet_netdev_priv_t *priv = netdev_priv(dev);

    memcpy(eth->h_source, dev->dev_addr, ETH_ALEN);
    skb->dev = dev;

    if(irqs_disabled()) {       /* tiatm is source */
        skb_queue_tail(&priv->mcfw_queue, skb);
        tasklet_schedule(&priv->mcfw_tasklet);
    } else {
        dev_queue_xmit(skb);
    }
}


#if 0
/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static void avmnet_mc_transmit(void           *privatedata,
                              int             sourceid __attribute__ ((unused)),
                              mcfw_portset    portset,
                              struct sk_buff *skb) {
    unsigned char instance_number = (unsigned char) (((unsigned int) privatedata) & 0xff);
    cpmac_instance_t *instance;
    struct net_device *p_dev;
    struct sk_buff *tskb;
    unsigned short vid;
    t_cpphy_switch_config *switch_config;

    assert(instance_number < AVM_CPMAC_MAX_INSTANCES);
    instance = &cpmac_global.instance[instance_number];

#   if defined(CONFIG_ARCH_PUMA5)
    p_dev = cpmac_puma_get_dev();
#   else /*--- #if defined(CONFIG_ARCH_PUMA5) ---*/
    p_dev = instance->dev;
    switch_config = &instance->mdio.switch_config;
#   endif /*--- #else ---*/ /*--- #if defined(CONFIG_ARCH_PUMA5) ---*/

    if(p_dev == 0) {
        dev_kfree_skb_any(skb);
        return;
    }

    /* MC streams go to the LAN */
    skb->uniq_id &= 0xffffff;
    skb->uniq_id |= (CPPHY_PRIO_QUEUE_LAN << 24);
    assert(portset <= AVM_CPMAC_MAX_PORTSET);
    vid = switch_config->map_portset_out[portset];
    vid = cpmac_global.instance[0].mdio.switch_config.map_portset_out[portset];
    if(vid != 0) {
        if((tskb = __vlan_put_tag(skb, vid)) != 0) {
            avmnet_dev_queue_xmit(p_dev, tskb, instance_number);
        } else {
            dev_kfree_skb_any(skb);
        }
    } else {
        avmnet_dev_queue_xmit(p_dev, skb, instance_number);
    }
}

#endif /*--- #if 0 ---*/

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static void avmnet_mc_transmit_single(void            *privatedata,
                                      int              sourceid __attribute__ ((unused)),
                                      unsigned         n __attribute__ ((unused)),
                                      mcfw_memberinfo *members __attribute__ ((unused)),
                                      struct sk_buff  *skb) {
    avmnet_device_t *avm_dev = (avmnet_device_t *) privatedata;

    if(skb == NULL) {
        AVMNET_ERR("[%s] skb == NULL!\n", __FUNCTION__);
        return;
    }

    avmnet_dev_queue_xmit(avm_dev, skb);
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int avmnet_mcfw_init(avmnet_device_t *avm_dev) {
    struct net_device *dev = avm_dev->device;
    avmnet_netdev_priv_t *priv = netdev_priv(dev);
    int err;

    priv->mcfw_used = 0;
    strncpy(priv->mcfw_netdrv.name, dev->name, MCFWDRV_NAMSIZ - 1);
    priv->mcfw_sourceid = 0;
    priv->mcfw_netdrv.netdriver_mc_transmit_single = avmnet_mc_transmit_single;
    priv->mcfw_netdrv.privatedata = (void *) avm_dev;

    skb_queue_head_init(&priv->mcfw_queue);
    tasklet_init(&priv->mcfw_tasklet, avmnet_mcfw_handle_tasklet, (unsigned long) avm_dev);

    err = mcfw_netdriver_register(&priv->mcfw_netdrv);
    if(err) {
        AVMNET_WARN("[%s] Could not register mcfw_netdriver!\n", __FUNCTION__);
        return err;
    }
    priv->mcfw_used = 1;
    priv->mcfw_sourceid = mcfw_multicast_forward_alloc_id(priv->mcfw_netdrv.name);
    AVMNET_INFO("[%s] Register multicast forward for '%s'\n", 
                __FUNCTION__, dev->name);

    return err;
}


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
#if 0
void avmnet_mcfw_exit(unsigned char instance_number) {
    cpmac_instance_t *instance = &cpmac_global.instance[instance_number];

    if(instance->mcfw_used) {
        instance->mcfw_used = 0;
        mcfw_netdriver_unregister(&instance->mcfw_netdrv);
        mcfw_multicast_forward_free_id(instance->mcfw_sourceid);
        tasklet_kill(&instance->mcfw_tasklet);
        DEB_INFO("[%s] done\n", __FUNCTION__);
    }
}
#endif /*--- #if 0 ---*/

#endif /*--- #if defined(CONFIG_IP_MULTICAST_FASTFORWARD) ---*/

