
/*------------------------------------------------------------------------------------------*\
 *   Copyright (C) 2011 AVM GmbH <fritzbox_info@avm.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
\*------------------------------------------------------------------------------------------*/

#include <linux/sched.h>
#include <linux/types.h>
#include <linux/list.h>
#include <linux/workqueue.h>

#include <avmnet_debug.h>
#include <avmnet_module.h>
#include <avmnet_config.h>

#ifdef CONFIG_AVM_PA
#include <linux/avm_pa.h>
#include <linux/avm_pa_hw.h>
#include <linux/avm_pa_intern.h>
#include <linux/avm_pa_ifx_multiplexer.h>
#endif

#include <net/ifx_ppa_ppe_hal.h>
#include "swi_ifx_common.h"

struct avm_pa_virt_rx_dev *ppa_virt_rx_devs[MAX_PPA_PUSHTHROUGH_DEVICES];
struct avm_pa_virt_tx_dev *ppa_virt_tx_devs[MAX_PPA_PUSHTHROUGH_DEVICES];
EXPORT_SYMBOL(ppa_virt_rx_devs);
EXPORT_SYMBOL(ppa_virt_tx_devs);
/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

unsigned int push_through_devs_supported_by_datapath = 0;
EXPORT_SYMBOL(push_through_devs_supported_by_datapath);

int (*ppa_pushthrough_hook)(uint32_t, struct sk_buff*, int32_t , uint32_t) = NULL;
EXPORT_SYMBOL(ppa_pushthrough_hook);

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

int check_if_avmnet_enables_ppa(void) {
	avmnet_module_t *this = avmnet_hw_config_entry->config;
    return !(this->initdata.swi.flags & SWI_DISABLE_LANTIQ_PPA);
}
EXPORT_SYMBOL(check_if_avmnet_enables_ppa);

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int avm_wan_mode = AVM_WAN_MODE_NOT_CONFIGURED;
EXPORT_SYMBOL(avm_wan_mode);

int g_dbg_datapath = DBG_ENABLE_MASK_ERR ;
module_param(g_dbg_datapath, int, S_IRUSR | S_IWUSR); //kernel_commandline: swi_ifx_common.g_dbg_datapath
EXPORT_SYMBOL(g_dbg_datapath);

int g_datapath_mod_count = 0;
EXPORT_SYMBOL(g_datapath_mod_count);

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

char common_proc_read_buff[COMMON_BUFF_LEN];
EXPORT_SYMBOL(common_proc_read_buff);

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

inline int avm_dev_to_mac_nr(avmnet_device_t *avm_dev)
{
    return avm_dev->mac_module->initdata.mac.mac_nr;
}
EXPORT_SYMBOL(avm_dev_to_mac_nr);

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

int proc_read_avm_ata_dev(char *page,
                         char **start __attribute__ ((unused)),
                         off_t off,
                         int count __attribute__ ((unused)),
                         int *eof,
                         void *data __attribute__ ((unused))){
    int len = 0;
    char avm_ata_dev_name[NETDEV_NAME_LEN];

    if ( avmnet_swi_ifx_common_get_ethwan_ata_devname( avm_ata_dev_name )){
        len += sprintf(page + off + len, "%s", avm_ata_dev_name);
    }

    if (len == 0)
        len += sprintf(page + off + len, "NONE");

    *eof = 1;

    return len;
}
/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int proc_write_avm_ata_dev(struct file *file __attribute__ ((unused)),
                          const char *buf,
                          unsigned long count,
                          void *data __attribute__ ((unused))) {
    int rlen;
    char avm_ata_dev_name[NETDEV_NAME_LEN];
    int len = count < NETDEV_NAME_LEN ? count :  NETDEV_NAME_LEN - 1;
    rlen = copy_from_user( avm_ata_dev_name, buf, len );
    avm_ata_dev_name[len] = 0;
#if defined(CONFIG_VR9)
    avmnet_swi_vr9_set_ethwan_dev_by_name( avm_ata_dev_name );
#endif

    return count;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int (*in_avm_ata_mode_hook)(void) = NULL;
EXPORT_SYMBOL(in_avm_ata_mode_hook);
/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

int in_avm_ata_mode(void) {
	// printk(KERN_ERR "[%s] hook =%pF\n", __FUNCTION__,in_avm_ata_mode_hook );
    if( !in_avm_ata_mode_hook ) return 0;
    return in_avm_ata_mode_hook();
}
EXPORT_SYMBOL(in_avm_ata_mode);
/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

int proc_read_avm_wan_mode(char *page,
                         char **start __attribute__ ((unused)),
                         off_t off,
                         int count __attribute__ ((unused)),
                         int *eof,
                         void *data __attribute__ ((unused))){
    int len = 0;

    switch(avm_wan_mode){
    case AVM_WAN_MODE_NOT_CONFIGURED:
        len += sprintf(page + off + len, "NONE");
        break;
    case AVM_WAN_MODE_ATM:
        len += sprintf(page + off + len, "ATM");
        break;
    case AVM_WAN_MODE_PTM:
        len += sprintf(page + off + len, "PTM");
        break;
    case AVM_WAN_MODE_ETH:
        len += sprintf(page + off + len, "ETH");
        break;
    case AVM_WAN_MODE_KDEV:
        len += sprintf(page + off + len, "KDEV");
        break;
    default:
        len += sprintf(page + off + len, "UNKNOWN");
        break;
    }

    *eof = 1;

    return len;
}
/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int proc_read_dbg(char *page,
                         char **start __attribute__ ((unused)),
                         off_t off,
                         int count __attribute__ ((unused)),
                         int *eof,
                         void *data __attribute__ ((unused))){
	int len = 0;

    len += sprintf(page + off + len, "error print        - %s\n", (g_dbg_datapath & DBG_ENABLE_MASK_ERR)              ? "enabled" : "disabled");
    len += sprintf(page + off + len, "debug print        - %s\n", (g_dbg_datapath & DBG_ENABLE_MASK_DEBUG_PRINT)      ? "enabled" : "disabled");
    len += sprintf(page + off + len, "assert             - %s\n", (g_dbg_datapath & DBG_ENABLE_MASK_ASSERT)           ? "enabled" : "disabled");
    len += sprintf(page + off + len, "dump rx skb        - %s\n", (g_dbg_datapath & DBG_ENABLE_MASK_DUMP_SKB_RX)      ? "enabled" : "disabled");
    len += sprintf(page + off + len, "dump tx skb        - %s\n", (g_dbg_datapath & DBG_ENABLE_MASK_DUMP_SKB_TX)      ? "enabled" : "disabled");
    len += sprintf(page + off + len, "dump flag header   - %s\n", (g_dbg_datapath & DBG_ENABLE_MASK_DUMP_FLAG_HEADER) ? "enabled" : "disabled");
    len += sprintf(page + off + len, "dump egress header - %s\n", (g_dbg_datapath & DBG_ENABLE_MASK_DUMP_EG_HEADER)   ? "enabled" : "disabled");
    len += sprintf(page + off + len, "dump init          - %s\n", (g_dbg_datapath & DBG_ENABLE_MASK_DUMP_INIT)        ? "enabled" : "disabled");
    len += sprintf(page + off + len, "mac swap           - %s\n", (g_dbg_datapath & DBG_ENABLE_MASK_MAC_SWAP)         ? "enabled" : "disabled");
    len += sprintf(page + off + len, "dump pmac header   - %s\n", (g_dbg_datapath & DBG_ENABLE_MASK_DUMP_PMAC_HEADER) ? "enabled" : "disabled");
    len += sprintf(page + off + len, "dump qos           - %s\n", (g_dbg_datapath & DBG_ENABLE_MASK_DUMP_QOS)         ? "enabled" : "disabled");
    len += sprintf(page + off + len, "debug dma          - %s\n", (g_dbg_datapath & DBG_ENABLE_MASK_DEBUG_DMA)        ? "enabled" : "disabled");
    len += sprintf(page + off + len, "dump atm tx        - %s\n", (g_dbg_datapath & DBG_ENABLE_MASK_DUMP_ATM_TX)      ? "enabled" : "disabled");
    len += sprintf(page + off + len, "dump atm rx        - %s\n", (g_dbg_datapath & DBG_ENABLE_MASK_DUMP_ATM_RX)      ? "enabled" : "disabled");
    len += sprintf(page + off + len, "debug mailbox      - %s\n", (g_dbg_datapath & DBG_ENABLE_MASK_DEBUG_MAILBOX)    ? "enabled" : "disabled");
    len += sprintf(page + off + len, "debug alignment    - %s\n", (g_dbg_datapath & DBG_ENABLE_MASK_DEBUG_ALIGNMENT)  ? "enabled" : "disabled");
    len += sprintf(page + off + len, "debug trace        - %s\n", (g_dbg_datapath & DBG_ENABLE_MASK_DEBUG_TRACE_PPA_DATA)  ? "enabled" : "disabled");
    len += sprintf(page + off + len, "debug vdev         - %s\n", (g_dbg_datapath & DBG_ENABLE_MASK_DEBUG_VDEV          )  ? "enabled" : "disabled");

#if defined(ENABLE_FW_MODULE_TEST) && ENABLE_FW_MODULE_TEST
    len += sprintf(page + off + len, "A5: FW Mode          - DSL WAN\n");
    len += sprintf(page + off + len, "A5: WAN Hash Algo    - %s\n", (ppe_wan_hash_algo) ? "Algo 1" : "Algo 0");
    len += sprintf(page + off + len, "A5: ACC Protocol     - %s\n", (acc_proto) ? "TCP" : "UDP");
    len += sprintf(page + off + len, "A5: MPoA Type        - EoA1\n");
#endif

#if defined(ENABLE_FW_MODULE_TEST) && ENABLE_FW_MODULE_TEST
    len += sprintf(page + off + len, "E5: ipv6             - %s\n", (ipv6_en)             ? "enabled"  : "disabled");
    len += sprintf(page + off + len, "E5: fw_mode          - %s\n", (ppe_fw_mode)        ? "mix_mode" : "normal_mode");
    len += sprintf(page + off + len, "E5: wan_hash_algo    - %s\n", (ppe_wan_hash_algo)   ? "Algo 1"   : "Algo 0");
    len += sprintf(page + off + len, "E5: acc_protocol     - %s\n", (acc_proto)        ? "TCP" : "UDP");
#endif

	*eof = 1;

	return len;
}

int proc_write_dbg(struct file *file __attribute__ ((unused)),
                          const char *buf,
                          unsigned long count,
                          void *data __attribute__ ((unused)))
{
    static const char *dbg_enable_mask_str[] = {
        " error print",
        " err",
        " debug print",
        " dbg",
        " assert",
        " assert",
        " dump rx skb",
        " rx",
        " dump tx skb",
        " tx",
        " dump flag header",
        " flag",
        " dump egress header",
        " egress",
        " dump init",
        " init",
        " mac swap",
        " swap",
        " dump pmac header",
        " pmac",
        " dump qos",
        " qos",
        " debug dma",
        " dma",
        " dump atm tx",
        " atmtx",
        " dump atm rx",
        " atmrx",
        " debug mailbox",
        " mailbox",
        " debug alignment",
        " alignment",
        " debug trace",
        " trace",
        " debug vdev",
        " vdev",
        " all"
    };
    static const int dbg_enable_mask_str_len[] = {
        12, 4,
        12, 4,
        7,  7,
        12, 3,
        12, 3,
        17, 5,
        19, 7,
        10, 5,
        9,  5,
        17, 5,
         9,	4,
        10,	4,
        12,	6,
        12,	6,
        14,	8,
        16, 10,
        12, 6,
        11, 5,
        4
    };
    u32 dbg_enable_mask[] = {
        DBG_ENABLE_MASK_ERR,
        DBG_ENABLE_MASK_DEBUG_PRINT,
        DBG_ENABLE_MASK_ASSERT,
        DBG_ENABLE_MASK_DUMP_SKB_RX,
        DBG_ENABLE_MASK_DUMP_SKB_TX,
        DBG_ENABLE_MASK_DUMP_FLAG_HEADER,
        DBG_ENABLE_MASK_DUMP_EG_HEADER,
        DBG_ENABLE_MASK_DUMP_INIT,
        DBG_ENABLE_MASK_MAC_SWAP,
        DBG_ENABLE_MASK_DUMP_PMAC_HEADER,
        DBG_ENABLE_MASK_DUMP_QOS,
        DBG_ENABLE_MASK_DEBUG_DMA,
        DBG_ENABLE_MASK_DUMP_ATM_TX,
        DBG_ENABLE_MASK_DUMP_ATM_RX,
        DBG_ENABLE_MASK_DEBUG_MAILBOX,
        DBG_ENABLE_MASK_DEBUG_ALIGNMENT,
        DBG_ENABLE_MASK_DEBUG_TRACE_PPA_DATA,
        DBG_ENABLE_MASK_DEBUG_VDEV,
        DBG_ENABLE_MASK_ALL & ~DBG_ENABLE_MASK_MAC_SWAP
    };

    char *p;

    int len, rlen;

    int f_enable = 0;
    unsigned int i;

    len = count < sizeof(common_proc_read_buff) ? count : sizeof(common_proc_read_buff) - 1;
    rlen = len - copy_from_user(common_proc_read_buff, buf, len);
    while ( rlen && common_proc_read_buff[rlen - 1] <= ' ' )
        rlen--;
    common_proc_read_buff[rlen] = 0;
    for ( p = common_proc_read_buff; *p && *p <= ' '; p++, rlen-- );
    if ( !*p )
    {
        return 0;
    }

  #if defined(ENABLE_FW_MODULE_TEST) && ENABLE_FW_MODULE_TEST
    if ( ifx_strincmp(p, "enable_ipv6", 11) == 0 ) {
        ipv6_en = 1;
        ppe_fw_mode = 0;
        acc_proto = 0;
        ppe_wan_hash_algo = 0;
        printk("Mode          : Normal Node\n");
        printk("Protocol      : UDP\n");
        printk("WAN HASH Algo : Algorithm 0\n");
        init_acc_tables();
    } else if ( ifx_strincmp(p, "disable_ipv6", 12) == 0 ) {
        ipv6_en = 0;
        ppe_fw_mode = 0;
        acc_proto = 0;
        ppe_wan_hash_algo = 0;
        printk("Mode          : Normal Node\n");
        printk("Protocol      : UDP\n");
        printk("WAN HASH Algo : Algorithm 0\n");
        init_acc_tables();
    } else if ( ifx_strincmp(p, "set_acc_udp", 11) == 0 ) {
        if (acc_proto != 0) {
            acc_proto = 0;
            init_acc_tables();
        }
    } else if ( ifx_strincmp(p, "set_acc_tcp", 11) == 0 ) {
        if (acc_proto != 1) {
            acc_proto = 1;
            init_acc_tables();
        }
    } else if ( ifx_strincmp(p, "set_mix_mode", 12) == 0 ) {
        if (ppe_fw_mode != 1) {
            ppe_fw_mode = 1;
            init_acc_tables();
        }
    } else if ( ifx_strincmp(p, "set_normal_mode", 15) == 0 ) {
        if (ppe_fw_mode != 0) {
            ppe_fw_mode = 0;
            init_acc_tables();
        }
    }
    else if ( ifx_strincmp(p, "set_algo_0", 10) == 0 ) {
        if (ppe_wan_hash_algo != 0) {
            ppe_wan_hash_algo = 0;
            init_acc_tables();
        }
    } else if ( ifx_strincmp(p, "set_algo_1", 10) == 0 ) {
        if (ppe_wan_hash_algo != 1) {
            ppe_wan_hash_algo = 1;
            init_acc_tables();
        }
    } else
  #endif
    if ( ifx_strincmp(p, "enable", 6) == 0 )
    {
        p += 6;
        f_enable = 1;
    }
    else if ( ifx_strincmp(p, "disable", 7) == 0 )
    {
        p += 7;
        f_enable = -1;
    }
    else if ( ifx_strincmp(p, "help", 4) == 0 || *p == '?' )
    {
        printk("echo <enable/disable> [err/dbg/assert/rx/tx/header/init/all] > /proc/eth/dbg\n");
  #if defined(ENABLE_FW_MODULE_TEST) && ENABLE_FW_MODULE_TEST
        printk("echo enable_ipv6/disable_ipv6\n");
        printk("echo set_mix_mode/set_normal_mode/set_algo_0/set_algo_1\n");
        printk("echo set_acc_udp/set_acc_tcp\n");
  #endif
    }

    if ( f_enable )
    {
        if ( *p == 0 )
        {
            if ( f_enable > 0 )
                g_dbg_datapath |= DBG_ENABLE_MASK_ALL & ~DBG_ENABLE_MASK_MAC_SWAP;
            else
                g_dbg_datapath &= ~DBG_ENABLE_MASK_ALL | DBG_ENABLE_MASK_MAC_SWAP;
        }
        else
        {
            do
            {
                for ( i = 0; i < NUM_ENTITY(dbg_enable_mask_str); i++ )
                    if ( ifx_strincmp(p, dbg_enable_mask_str[i], dbg_enable_mask_str_len[i]) == 0 )
                    {
                        if ( f_enable > 0 ){
                            g_dbg_datapath |= dbg_enable_mask[i >> 1];
                            printk(KERN_ERR "enable %d \n", i);
                        }
                        else
                            g_dbg_datapath &= ~dbg_enable_mask[i >> 1];
                        p += dbg_enable_mask_str_len[i];
                        break;
                    }
            } while ( i < NUM_ENTITY(dbg_enable_mask_str) );
        }
    }

    return count;
}


struct proc_dir_entry *g_eth_proc_dir = NULL;
EXPORT_SYMBOL(g_eth_proc_dir);

void ifx_proc_file_create(void) {
	struct proc_dir_entry *res = NULL;
	g_eth_proc_dir = proc_mkdir("eth", NULL);

	res = create_proc_entry("avm_wan_mode", 0, g_eth_proc_dir);
	if (res) {
		res->read_proc = proc_read_avm_wan_mode;
	}

	res = create_proc_entry("avm_ata_dev", 0, g_eth_proc_dir);
	if (res) {
		res->read_proc = proc_read_avm_ata_dev;
		res->write_proc = proc_write_avm_ata_dev;
	}

#if defined(ENABLE_DBG_PROC) && ENABLE_DBG_PROC
	res = create_proc_read_entry("dbg", 0, g_eth_proc_dir, proc_read_dbg, NULL);
	if (res)
		res->write_proc = proc_write_dbg;
#endif

#if defined(DEBUG_PP32_PROC) && DEBUG_PP32_PROC
	res = create_proc_entry("pp32", 0, g_eth_proc_dir);
	if (res) {
		res->read_proc = proc_read_pp32;
		res->write_proc = proc_write_pp32;
	}
#endif

}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

#if defined(DEBUG_DUMP_SKB) && DEBUG_DUMP_SKB
void dump_skb(struct sk_buff *skb, u32 len, char *title, int port, int ch, int is_tx, int enforce) {
    unsigned int i;

    if ( !enforce && !(g_dbg_datapath & (is_tx ? DBG_ENABLE_MASK_DUMP_SKB_TX : DBG_ENABLE_MASK_DUMP_SKB_RX)) )
        return;

    if ( skb->len < len )
        len = skb->len;

    if ( len > ETH_PKT_BUF_SIZE )
    {
        printk("too big data length: skb = %08x, skb->data = %08x, skb->len = %d\n", (u32)skb, (u32)skb->data, skb->len);
        return;
    }

    if ( ch >= 0 )
        printk("%s (port %d, ch %d)\n", title, port, ch);
    else
        printk("%s\n", title);
    printk("  skb->data = %08X, skb->tail = %08X, skb->len = %d\n", (u32)skb->data, (u32)skb->tail, (int)skb->len);
    for ( i = 1; i <= len; i++ )
    {
        if ( i % 16 == 1 )
            printk("  %4d:", i - 1);
        printk(" %02X", (int)(*((char*)skb->data + i - 1) & 0xFF));
        if ( i % 16 == 0 )
            printk("\n");
    }
    if ( (i - 1) % 16 != 0 )
        printk("\n");
}
EXPORT_SYMBOL(dump_skb);
#endif

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
avmnet_device_t *mac_nr_to_avm_dev(int mac_nr)
{
    static int optimized = 0;
    static avmnet_device_t *lookup_table[MAX_ETH_INF];
    int i;
    unsigned long flags;

    if( unlikely( !optimized ) ){
        avmnet_device_t **avm_devices;
        local_irq_save(flags);
         avm_devices = avmnet_hw_config_entry->avm_devices;

        for (i = 0; i < MAX_ETH_INF; i++)
        	lookup_table[i] = NULL;

        for(i = 0; i < avmnet_hw_config_entry->nr_avm_devices; i++){
            if(avm_devices[i]->mac_module){
                if(avm_devices[i]->mac_module->initdata.mac.mac_nr < MAX_ETH_INF)
                    lookup_table[avm_devices[i]->mac_module->initdata.mac.mac_nr] =
                            avm_devices[i];
                else
                    BUG();
            }
        }
        AVMNET_INFO("[%s] set up lookup_table\n", __FUNCTION__);
        optimized = 1;
        local_irq_restore(flags);
    }

    if (mac_nr < MAX_ETH_INF)
        return lookup_table[mac_nr];
    else
        return NULL;

}
EXPORT_SYMBOL(mac_nr_to_avm_dev);

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/

/*
 * setup wan dev by name in AVM_PA flags, returns mac_wan_mask
 */
int avmnet_swi_ifx_common_set_ethwan_dev_by_name( const char *devname ) {
	int wan_phy_mask = 0;
	int i;

	if (!devname)
	    return 0;

	AVMNET_ERR("[%s] %s\n", __func__, devname);

    for(i = 0; i < avmnet_hw_config_entry->nr_avm_devices; i++){
        struct avm_pa_pid_hwinfo *hw_info = NULL;
        if (avmnet_hw_config_entry->avm_devices[i]->device)
                hw_info = avm_pa_pid_get_hwinfo(avmnet_hw_config_entry->avm_devices[i]->device->avm_pa.devinfo.pid_handle);

    	if (avmnet_hw_config_entry->avm_devices[i]->device_name &&
    	        ( strncmp( avmnet_hw_config_entry->avm_devices[i]->device_name, devname, NETDEV_NAME_LEN ) == 0)) {
    	    if ( avmnet_hw_config_entry->avm_devices[i]->mac_module ) {

    	        // add if to wan_phy_mask
    		    wan_phy_mask |= ( 1 << avmnet_hw_config_entry->avm_devices[i]->mac_module->initdata.mac.mac_nr );

    		    //set ATA FLAG in AVM_PA HWINFO
    		    if (hw_info){
                	AVMNET_ERR("[%s] set hw info ATA\n", __func__);
                    hw_info->flags |= AVMNET_DEVICE_IFXPPA_ETH_WAN;
    		    }
    	    }
    	} else {
    		    //reset ATA FLAG in AVM_PA HWINFO
    		    if ( avmnet_hw_config_entry->avm_devices[i]->mac_module
    		            && hw_info
    		            && (hw_info->flags & AVMNET_DEVICE_IFXPPA_ETH_WAN )){
                    AVMNET_ERR("[%s] reset hw info ATA\n", __func__);
                    hw_info->flags &= ~AVMNET_DEVICE_IFXPPA_ETH_WAN;
    		    }
    	}
    }
    return wan_phy_mask;
}
EXPORT_SYMBOL(avmnet_swi_ifx_common_set_ethwan_dev_by_name);

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int avmnet_swi_ifx_common_get_ethwan_ata_devname( char *devname ) {
	int i;
	int wan_phy_mask = 0;

    for(i = 0; i < avmnet_hw_config_entry->nr_avm_devices; i++){
        struct avm_pa_pid_hwinfo *hw_info = NULL;

   	    if ( !avmnet_hw_config_entry->avm_devices[i]->mac_module )
   	        continue;

   	    if (avmnet_hw_config_entry->avm_devices[i]->device)
            hw_info = avm_pa_pid_get_hwinfo(avmnet_hw_config_entry->avm_devices[i]->device->avm_pa.devinfo.pid_handle);

 		if (hw_info && (hw_info->flags & AVMNET_DEVICE_IFXPPA_ETH_WAN)){

   		    wan_phy_mask = ( 1 << avmnet_hw_config_entry->avm_devices[i]->mac_module->initdata.mac.mac_nr );

 		    if ( avmnet_hw_config_entry->avm_devices[i]->device_name )
                snprintf(devname, NETDEV_NAME_LEN , "%s", avmnet_hw_config_entry->avm_devices[i]->device_name );

            return wan_phy_mask;
 		}
    }

    *devname = 0;
    return 0;
}
EXPORT_SYMBOL(avmnet_swi_ifx_common_get_ethwan_ata_devname);


/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
#define PUSHTHROUGH_DATA_PATH_NOT_READY   2
#define PUSHTHROUGH_DEV_UNKNOWN           1
#define PACKET_SENT                       0
#define PACKET_DROPPED                   -1

/*------------------------------------------------------------------------------------------*\
 * Returns:
 *  0 <= Packet was absorbed by HW-PPA (sent or dropped)
 *  0 >  Packet has to be processed by AVM_PA
\*------------------------------------------------------------------------------------------*/
int ifx_ppa_try_to_accelerate( avm_pid_handle pid_handle, struct sk_buff *skb ){
    struct avm_pa_pid_hwinfo *hwinfo = NULL;
    struct avm_pa_virt_rx_dev *virtdev = NULL;
    int hw_dev_id;

    hwinfo = avm_pa_pid_get_hwinfo( pid_handle );
    if (hwinfo)
        virtdev = hwinfo->virt_rx_dev;

    if ( !virtdev
        || ( virtdev->hw_dev_id > push_through_devs_supported_by_datapath)
        || ( virtdev != ppa_virt_rx_devs[ virtdev->hw_dev_id ] )){

            dbg( "hw_dev_id is not registerd for avm_pid_handle=%d", pid_handle );
            if (virtdev)
                virtdev->hw_pkt_try_to_acc_virt_chan_not_ready ++;

            return PUSHTHROUGH_DEV_UNKNOWN;
    }
    hw_dev_id = virtdev->hw_dev_id;

    dbg_trace_ppa_data("virt_dev_id %d",hw_dev_id );
    if ( ppa_pushthrough_hook ){
        int ret_val;

        dbg_trace_ppa_data("virt_dev_id %d",hw_dev_id );
        /* datapath module is ready for pushthrough
         * returns
         *  0  := packet sent
         *  -1 := packet dropped
         */
        ret_val = ppa_pushthrough_hook( hw_dev_id, skb, 0, 0 );
        if (ret_val < 0 )
            virtdev->hw_pkt_try_to_acc_dropped++;
        else
            virtdev->hw_pkt_try_to_acc++;

        return ret_val;

    } else {

        virtdev->hw_pkt_try_to_acc_virt_chan_not_ready ++;
        // datapath module is not ready for pushthrough, call rx_second_part from device
        return PUSHTHROUGH_DATA_PATH_NOT_READY;

    }

}

EXPORT_SYMBOL(ifx_ppa_try_to_accelerate);
/*------------------------------------------------------------------------------------------*\
 * Returns
 *  ( virt_dev_id >= 0 ): success
 *  (               -1 ): failed
\*------------------------------------------------------------------------------------------*/

int ifx_ppa_alloc_virtual_rx_device( avm_pid_handle pid_handle ){
    struct avm_pa_pid_hwinfo *hwinfo;
    struct avm_pa_virt_rx_dev *virtdev;
    int i;

    virtdev = kmalloc(sizeof(struct avm_pa_virt_rx_dev), GFP_KERNEL);

    if (!virtdev)
        return ALLOC_VIRT_DEV_FAILED;

    memset(virtdev, 0, sizeof(struct avm_pa_virt_rx_dev));
    virtdev->pid_handle = pid_handle;

    dbg("try to alloc virtual rx for pid %d", pid_handle);
    for (i=0; i < MAX_PPA_PUSHTHROUGH_DEVICES; i++){
        if ( ppa_virt_rx_devs[i] == NULL ){
            ppa_virt_rx_devs[i] = virtdev;
            virtdev->hw_dev_id = i;
            hwinfo = avm_pa_pid_get_hwinfo( pid_handle );

            if ( !hwinfo ){
                struct avm_pa_pid_hwinfo tmp_hwinfo = {0};
                dbg("hwinfo not set yet -> create new hwinfo for pid=%d", pid_handle);
                avm_pa_pid_set_hwinfo( pid_handle, &tmp_hwinfo );
                hwinfo = avm_pa_pid_get_hwinfo( pid_handle );
            }

            hwinfo->flags |= AVMNET_DEVICE_IFXPPA_VIRT_RX;
            hwinfo->virt_rx_dev = virtdev;
            return i;
        }
    }

    err("alloc virtual rx for pid %d FAILED ", pid_handle);
    if (virtdev)
        kfree(virtdev);
    return ALLOC_VIRT_DEV_FAILED;
}
EXPORT_SYMBOL(ifx_ppa_alloc_virtual_rx_device);


/*------------------------------------------------------------------------------------------*\
 * Returns
 *  ( virt_dev_id >= 0 ): success
 *  (               -1 ): failed
\*------------------------------------------------------------------------------------------*/
int ifx_ppa_alloc_virtual_tx_device(avm_pid_handle pid_handle){
    struct avm_pa_pid_hwinfo *hwinfo;
    struct avm_pa_virt_tx_dev *virtdev;
    int i;

    virtdev = kmalloc(sizeof(struct avm_pa_virt_tx_dev), GFP_KERNEL);

    if (!virtdev)
        return ALLOC_VIRT_DEV_FAILED;

    memset(virtdev, 0, sizeof(struct avm_pa_virt_tx_dev));
    virtdev->pid_handle = pid_handle;

    dbg("try to alloc virtual tx for pid %d", pid_handle);
    for (i=0; i < MAX_PPA_PUSHTHROUGH_DEVICES; i++){
        if ( ppa_virt_tx_devs[i] == NULL ){
            ppa_virt_tx_devs[i] = virtdev;
            virtdev->hw_dev_id = i;
            hwinfo = avm_pa_pid_get_hwinfo( pid_handle );

            if ( !hwinfo ){
                struct avm_pa_pid_hwinfo tmp_hwinfo = {0};
                dbg("hwinfo not set yet -> create new hwinfo for pid=%d", pid_handle);
                avm_pa_pid_set_hwinfo( pid_handle, &tmp_hwinfo );
                hwinfo = avm_pa_pid_get_hwinfo( pid_handle );
            }

            hwinfo->flags |= AVMNET_DEVICE_IFXPPA_VIRT_TX;
            hwinfo->virt_tx_dev = virtdev;
            return i;
        }
    }

    err("alloc virtual tx for pid %d FAILED ", pid_handle);
    if (virtdev)
        kfree(virtdev);
    return ALLOC_VIRT_DEV_FAILED;
}
EXPORT_SYMBOL(ifx_ppa_alloc_virtual_tx_device);


/*------------------------------------------------------------------------------------------*\
 * Returns
 *  ( virt_dev_id >= 0 ): success
 *  (               -1 ): failed
\*------------------------------------------------------------------------------------------*/
int ifx_ppa_free_virtual_tx_device(avm_pid_handle pid_handle){

    struct avm_pa_pid_hwinfo *hwinfo;
    struct avm_pa_virt_rx_dev *virtdev;
    int virt_dev_id;

    hwinfo = avm_pa_pid_get_hwinfo( pid_handle );
    if (! hwinfo )
        return FREE_VIRT_DEV_FAILED;

    virtdev = hwinfo->virt_tx_dev;
    if (! virtdev )
        return FREE_VIRT_DEV_FAILED;

    virt_dev_id = virtdev->hw_dev_id;
    if ( virt_dev_id < MAX_PPA_PUSHTHROUGH_DEVICES ){
        ppa_virt_tx_devs[ virt_dev_id ] = NULL;
    }
    hwinfo->virt_tx_dev = NULL;
    hwinfo->flags &= ~AVMNET_DEVICE_IFXPPA_VIRT_TX;

    kfree(virtdev);
    return virt_dev_id;
}


/*------------------------------------------------------------------------------------------*\
 * Returns
 *  ( virt_dev_id >= 0 ): success
 *  (               -1 ): failed
\*------------------------------------------------------------------------------------------*/
int ifx_ppa_free_virtual_rx_device(avm_pid_handle pid_handle){

    struct avm_pa_pid_hwinfo *hwinfo;
    struct avm_pa_virt_rx_dev *virtdev;
    int virt_dev_id;

    hwinfo = avm_pa_pid_get_hwinfo( pid_handle );
    if (! hwinfo )
        return FREE_VIRT_DEV_FAILED;

    virtdev = hwinfo->virt_rx_dev;
    if (! virtdev )
        return FREE_VIRT_DEV_FAILED;

    virt_dev_id = virtdev->hw_dev_id;
    if ( virt_dev_id < MAX_PPA_PUSHTHROUGH_DEVICES ){
        ppa_virt_rx_devs[ virt_dev_id ] = NULL;
    }
    hwinfo->virt_rx_dev = NULL;
    hwinfo->flags &= ~AVMNET_DEVICE_IFXPPA_VIRT_RX;

    kfree(virtdev);
    return virt_dev_id;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int ifx_ppa_show_virtual_devices(char *buff, unsigned int buff_len){
    int i;
    unsigned int used_buffer_size = 0;
    if (used_buffer_size < buff_len)
        used_buffer_size += snprintf(buff + used_buffer_size, buff_len - used_buffer_size, "push_through_devs_supported_by_datapath: %u\n", push_through_devs_supported_by_datapath);
    for (i=0; i < MAX_PPA_PUSHTHROUGH_DEVICES; i++){
        if ( ppa_virt_rx_devs[i] ){
            if (used_buffer_size < buff_len)
                used_buffer_size += snprintf(buff + used_buffer_size, buff_len - used_buffer_size, "virt_rx_dev[%d]: \n", i);
            if (used_buffer_size < buff_len)
                used_buffer_size += snprintf(buff + used_buffer_size, buff_len - used_buffer_size, "   hw_dev_id[rx]        : %lu \n", ppa_virt_rx_devs[i]->hw_dev_id );
            if (used_buffer_size < buff_len)
                used_buffer_size += snprintf(buff + used_buffer_size, buff_len - used_buffer_size, "   ppa_dev_id           : %lu \n", ppa_virt_rx_devs[i]->hw_dev_id + IFX_VIRT_DEV_OFFSET );
            if (used_buffer_size < buff_len)
                used_buffer_size += snprintf(buff + used_buffer_size, buff_len - used_buffer_size, "   slow_cnt             : %lu \n", ppa_virt_rx_devs[i]->hw_pkt_slow_cnt );
            if (used_buffer_size < buff_len)
                used_buffer_size += snprintf(buff + used_buffer_size, buff_len - used_buffer_size, "   try_to_acc           : %lu \n", ppa_virt_rx_devs[i]->hw_pkt_try_to_acc );
            if (used_buffer_size < buff_len)
                used_buffer_size += snprintf(buff + used_buffer_size, buff_len - used_buffer_size, "   try_to_acc_dropped   : %lu \n", ppa_virt_rx_devs[i]->hw_pkt_try_to_acc_dropped );
            if (used_buffer_size < buff_len)
                used_buffer_size += snprintf(buff + used_buffer_size, buff_len - used_buffer_size, "   try_to_acc_not_ready : %lu \n", ppa_virt_rx_devs[i]->hw_pkt_try_to_acc_virt_chan_not_ready );
        }
    }
    for (i=0; i < MAX_PPA_PUSHTHROUGH_DEVICES; i++){
        if ( ppa_virt_tx_devs[i] ){
            if (used_buffer_size < buff_len)
                used_buffer_size += snprintf(buff + used_buffer_size, buff_len - used_buffer_size, "virt_tx_dev[%d]: \n", i);
            if (used_buffer_size < buff_len)
                used_buffer_size += snprintf(buff + used_buffer_size, buff_len - used_buffer_size, "   hw_dev_id[tx]       : %lu \n", ppa_virt_tx_devs[i]->hw_dev_id );
            if (used_buffer_size < buff_len)
                used_buffer_size += snprintf(buff + used_buffer_size, buff_len - used_buffer_size, "   ppa_dev_id          : %lu \n", ppa_virt_tx_devs[i]->hw_dev_id + IFX_VIRT_DEV_OFFSET );
            if (used_buffer_size < buff_len)
                used_buffer_size += snprintf(buff + used_buffer_size, buff_len - used_buffer_size, "   hw_pkt_tx           : %lu \n", ppa_virt_tx_devs[i]->hw_pkt_tx);
            if (used_buffer_size < buff_len)
                used_buffer_size += snprintf(buff + used_buffer_size, buff_len - used_buffer_size, "   hw_pkt_slookup_fail : %lu \n", ppa_virt_tx_devs[i]->hw_pkt_tx_session_lookup_failed);
        }
    }
    return used_buffer_size;
}
EXPORT_SYMBOL(ifx_ppa_show_virtual_devices);

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/



